from sys import stdout

from beartype import beartype
from loguru import logger

from utilities.logging import LogLevel
from utilities.pathlib import PathLike


@beartype
def setup_loguru(*, name: PathLike = "log") -> None:
    """Set up `loguru` with the standard format strings."""
    logger.remove()

    fmt = (
        "<green>{time:YYYY-MM-DD}</green>"
        " "
        "<bold><green>{time:HH:mm:ss}</green></bold>"
        "."
        "{time:SSS}"
        "  "
        "<cyan>{process.name}</cyan>-{process.id}"
        "  "
        "<green>{name}</green>-<cyan>{function}</cyan>"
        "  "
        "<bold><level>{level.name}</level></bold>"
        "\n"
        "{message}"
    )

    _ = logger.add(stdout, level=LogLevel.INFO, format=fmt, enqueue=True)
    _ = logger.add(
        name,
        level=LogLevel.DEBUG,
        enqueue=True,
        rotation="10 MB",
        retention="1 week",
    )



from collections.abc import Iterable
from collections.abc import Iterator
from pathlib import Path
from typing import Any
from typing import Literal
from typing import Optional
from typing import TypeVar
from typing import Union
from typing import cast
from typing import overload

from beartype import beartype
from luigi import Target
from luigi import Task
from luigi import build as _build
from luigi.interface import LuigiRunResult
from luigi.notifications import smtp
from luigi.parameter import MissingParameterException
from luigi.task import Register
from luigi.task import flatten

from utilities.logging import LogLevel
from utilities.pathlib import PathLike


class PathTarget(Target):
    """A local target whose `path` attribute is a Pathlib instance."""

    @beartype
    def __init__(self, path: PathLike, /) -> None:
        super().__init__()
        self.path = Path(path)

    @beartype
    def exists(self) -> bool:
        """Check if the target exists."""
        return self.path.exists()


@overload
def build(
    task: Iterable[Task],
    /,
    *,
    detailed_summary: Literal[False] = False,
    local_scheduler: bool = False,
    log_level: Optional[LogLevel] = None,
    workers: Optional[int] = None,
) -> bool:
    ...


@overload
def build(
    task: Iterable[Task],
    /,
    *,
    detailed_summary: Literal[True],
    local_scheduler: bool = False,
    log_level: Optional[LogLevel] = None,
    workers: Optional[int] = None,
) -> LuigiRunResult:
    ...


@beartype
def build(
    task: Iterable[Task],
    /,
    *,
    detailed_summary: bool = False,
    local_scheduler: bool = False,
    log_level: Optional[LogLevel] = None,
    workers: Optional[int] = None,
) -> Union[bool, LuigiRunResult]:
    """Build a set of tasks."""
    return _build(
        task,
        detailed_summary=detailed_summary,
        local_scheduler=local_scheduler,
        **({} if log_level is None else {"log_level": log_level}),
        **({} if workers is None else {"workers": workers}),
    )


_Task = TypeVar("_Task", bound=Task)


@beartype
def clone(task: Task, cls: type[_Task], /, **kwargs: Any) -> _Task:
    """Clone a task."""
    return cast(_Task, task.clone(cls, **kwargs))


@overload
def get_dependencies_downstream(
    task: Task,
    /,
    *,
    cls: type[_Task],
    recursive: bool = False,
) -> frozenset[_Task]:
    ...


@overload
def get_dependencies_downstream(
    task: Task,
    /,
    *,
    cls: None = None,
    recursive: bool = False,
) -> frozenset[Task]:
    ...


@beartype
def get_dependencies_downstream(
    task: Task,
    /,
    *,
    cls: Optional[type[Task]] = None,
    recursive: bool = False,
) -> frozenset[Task]:
    """Get the downstream dependencies of a task."""
    return frozenset(
        _yield_dependencies_downstream(task, cls=cls, recursive=recursive),
    )


@beartype
def _yield_dependencies_downstream(
    task: Task,
    /,
    *,
    cls: Optional[type[Task]] = None,
    recursive: bool = False,
) -> Iterator[Task]:
    for task_cls in cast(Iterable[type[Task]], get_task_classes(cls=cls)):
        try:
            cloned = clone(task, task_cls)
        except (MissingParameterException, TypeError):
            pass
        else:
            if task in get_dependencies_upstream(cloned, recursive=recursive):
                yield cloned
                if recursive:
                    yield from get_dependencies_downstream(
                        cloned,
                        recursive=recursive,
                    )


@beartype
def get_dependencies_upstream(
    task: Task,
    /,
    *,
    recursive: bool = False,
) -> frozenset[Task]:
    """Get the upstream dependencies of a task."""
    return frozenset(_yield_dependencies_upstream(task, recursive=recursive))


@beartype
def _yield_dependencies_upstream(
    task: Task,
    /,
    *,
    recursive: bool = False,
) -> Iterator[Task]:
    for t in cast(Iterable[Task], flatten(task.requires())):
        yield t
        if recursive:
            yield from get_dependencies_upstream(t, recursive=recursive)


@overload
def get_task_classes(*, cls: type[_Task]) -> frozenset[type[_Task]]:
    ...


@overload
def get_task_classes(*, cls: None = None) -> frozenset[type[Task]]:
    ...


@beartype
def get_task_classes(
    *,
    cls: Optional[type[_Task]] = None,
) -> frozenset[type[_Task]]:
    """Yield the task classes. Optionally filter down."""
    return frozenset(_yield_task_classes(cls=cls))


@beartype
def _yield_task_classes(
    *,
    cls: Optional[type[_Task]] = None,
) -> Iterator[type[_Task]]:
    """Yield the task classes. Optionally filter down."""
    for name in Register.task_names():
        task_cls = Register.get_task_cls(name)
        if (
            (cls is None)
            or ((cls is not task_cls) and issubclass(task_cls, cls))
        ) and (task_cls is not smtp):
            yield cast(type[_Task], task_cls)



from typing import Any
from typing import Optional

from hypothesis.strategies import DrawFn
from hypothesis.strategies import composite
from sqlalchemy import MetaData
from sqlalchemy.engine import Engine

from utilities.hypothesis import temp_paths
from utilities.sqlalchemy import create_engine


@composite
def sqlite_engines(
    _draw: DrawFn,
    /,
    *,
    metadata: Optional[MetaData] = None,
    base: Any = None,
) -> Engine:
    """Strategy for generating SQLite engines."""
    temp_path = _draw(temp_paths())
    path = temp_path.joinpath("db.sqlite")
    engine = create_engine("sqlite", database=path.as_posix())
    if metadata is not None:
        metadata.create_all(engine)
    if base is not None:
        base.metadata.create_all(engine)

    # attach temp_path to the engine, so as to keep it alive
    engine.temp_path = temp_path  # type: ignore[]

    return engine



from pytest import raises

from utilities.typing import NeverError
from utilities.typing import never


class TestNever:
    def test_main(self) -> None:
        with raises(NeverError):
            never(None)  # type: ignore[]



from typing import Any
from typing import cast

from hypothesis import given
from luigi import Task
from sqlalchemy import Column
from sqlalchemy import Integer
from sqlalchemy.engine import Engine
from sqlalchemy.orm import declarative_base

from utilities.hypothesis.luigi import task_namespaces
from utilities.hypothesis.sqlalchemy import sqlite_engines
from utilities.luigi.sqlalchemy import EngineParameter
from utilities.luigi.sqlalchemy import TableParameter


class TestEngineParameter:
    @given(namespace=task_namespaces(), engine=sqlite_engines())
    def test_main(self, namespace: str, engine: Engine) -> None:
        class Example(Task):
            task_namespace = namespace

            engine = EngineParameter()

        _ = Example(engine)


class TestTableParameter:
    @given(namespace=task_namespaces())
    def test_main(self, namespace: str) -> None:
        class ExampleTask(Task):
            task_namespace = namespace

            table = TableParameter()

        class ExampleTable(cast(Any, declarative_base())):
            __tablename__ = "example"

            id_ = Column(Integer, primary_key=True)

        _ = ExampleTask(ExampleTable)



from dataclasses import dataclass
from pathlib import Path

from beartype import beartype


@beartype
@dataclass(frozen=True)
class Config:
    """Settings for the `rotate_logs` script."""

    path: Path = Path.cwd()
    extension: str = "log"
    size: int = int(100 * 1024)
    keep: int = 3
    dry_run: bool = False


@beartype
@dataclass(frozen=True)
class Item:
    """A log file, identified."""

    path: Path
    head: Path
    num: int = 0



import datetime as dt
from collections.abc import Iterator
from decimal import Decimal
from typing import Any

from beartype import beartype
from pandas import DataFrame
from pandas import NaT
from pandas import Series
from sqlalchemy.engine import Engine

from utilities.numpy import datetime64ns
from utilities.numpy import has_dtype
from utilities.pandas import Int64
from utilities.pandas import boolean
from utilities.pandas import string
from utilities.pandas import timestamp_to_date
from utilities.pandas import timestamp_to_datetime
from utilities.sqlalchemy import get_column_names
from utilities.sqlalchemy import get_columns
from utilities.sqlalchemy import get_table


@beartype
def insert_dataframe(
    df: DataFrame,
    table_or_model: Any,
    engine: Engine,
    /,
) -> None:
    """Insert a DataFrame into a database."""
    names = get_column_names(table_or_model)
    columns = get_columns(table_or_model)
    all_names_to_columns = dict(zip(names, columns))
    names_to_columns = {
        name: col
        for name, col in all_names_to_columns.items()
        if (df.columns == name).any()
    }
    nativized = (
        _nativize_column(df[name], column)
        for name, column in names_to_columns.items()
    )
    rows = zip(*nativized)
    dicts = [dict(zip(names_to_columns, row)) for row in rows]
    if len(dicts) >= 1:
        table = get_table(table_or_model)
        with engine.begin() as conn:
            _ = conn.execute(table.insert(), list(dicts))


@beartype
def _nativize_column(series: Series, column: Any, /) -> Iterator[Any]:
    py_type = column.type.python_type
    as_list = series.tolist()
    if (
        (has_dtype(series, (bool, boolean)) and (py_type in {bool, int}))
        or (has_dtype(series, float) and (py_type is float))
        or (has_dtype(series, (int, Int64)) and (py_type is int))
        or (has_dtype(series, string) and (py_type is str))
    ):
        values = as_list
    elif has_dtype(series, datetime64ns) and (py_type is dt.date):
        values = [None if t is NaT else timestamp_to_date(t) for t in as_list]
    elif has_dtype(series, datetime64ns) and (py_type is dt.datetime):
        values = [
            None if t is NaT else timestamp_to_datetime(t) for t in as_list
        ]
    else:
        msg = f"Invalid types: {series}, {py_type}"
        raise TypeError(msg)
    for is_null, native in zip(series.isna(), values):
        yield None if is_null else native


@beartype
def read_dataframe(sel: Any, engine: Engine, /) -> DataFrame:
    """Read a table from a database into a DataFrame."""
    with engine.begin() as conn:
        rows = conn.execute(sel).all()
    sel = {col.name: _get_dtype(col) for col in sel.selected_columns}
    return DataFrame(rows, columns=list(sel)).astype(sel)


@beartype
def _get_dtype(column: Any, /) -> Any:
    py_type = column.type.python_type
    if py_type is bool:
        return boolean
    if (py_type is float) or (py_type is Decimal):
        return float
    if py_type is int:
        return Int64
    if py_type is str:
        return string
    if issubclass(py_type, dt.date):
        return datetime64ns
    msg = f"Invalid type: {py_type}"  # pragma: no cover
    raise TypeError(msg)  # pragma: no cover



from enum import Enum
from typing import Any
from typing import TypeVar
from typing import Union

from beartype import beartype


class StrEnum(str, Enum):
    """An enum whose elements are themselves strings."""

    @staticmethod
    @beartype
    def _generate_next_value_(
        name: str,
        start: Any,
        count: int,
        last_values: Any,
    ) -> str:
        _ = start, count, last_values
        return name


_E = TypeVar("_E", bound=Enum)


@beartype
def parse_enum(enum: type[_E], x: Union[_E, str], /) -> _E:
    """Parse a string into the enum."""
    return enum[x] if isinstance(x, str) else x



from random import choice
from typing import Any
from typing import cast

from attrs import define
from attrs import fields
from beartype import beartype
from beartype.door import die_if_unbearable


@define
class AttrsBase:
    """Base class for `attrs` class which applies `beartype` checking."""

    @beartype
    def __attrs_post_init__(self) -> None:
        all_fields = fields(cast(Any, type(self)))
        try:
            field = choice(all_fields)
        except IndexError:
            pass
        else:
            die_if_unbearable(getattr(self, field.name), field.type)



from typing import Any
from typing import cast

from attrs import define
from attrs import fields
from beartype.door import die_if_unbearable
from beartype.roar import BeartypeAbbyHintViolation
from pytest import raises

from utilities.attrs import AttrsBase
from utilities.timer import Timer


class TestAttrsBase:
    def test_main(self) -> None:
        @define
        class Example(AttrsBase):
            x: list[int]

        with raises(BeartypeAbbyHintViolation):
            _ = Example(["0"])  # type: ignore[]

    def test_empty(self) -> None:
        @define
        class Example(AttrsBase):
            pass

        _ = Example()

    def test_speed(self) -> None:
        @define
        class Example:
            x: int
            y: int

        @define
        class Full:
            x: int
            y: int

            def __attrs_post_init__(self) -> None:
                for field in fields(cast(Any, type(self))):
                    die_if_unbearable(getattr(self, field.name), field.type)

        n = int(1e4)
        with Timer() as timer1:
            for _ in range(n):
                _ = Example(0, 0)
        with Timer() as timer2:
            for _ in range(n):
                _ = Full(0, 0)
        assert timer1 < timer2



from typing import Any

from beartype import beartype
from hypothesis.strategies import composite

from utilities.hypothesis import temp_paths


@composite
@beartype
def task_namespaces(_draw: Any, /) -> str:
    """Strategy for generating task namespaces."""
    path = _draw(temp_paths())
    return path.name



from os import environ

from utilities.pytest import is_pytest


class TestIsPytest:
    def test_main(self) -> None:
        assert is_pytest()

    def test_disable(self) -> None:
        del environ["PYTEST_CURRENT_TEST"]
        assert not is_pytest()



from smtplib import SMTPServerDisconnected

from pytest import raises

from utilities.airium import yield_airium
from utilities.email import InvalidContentsError
from utilities.email import send_email


class TestSendEmail:
    def test_main(self) -> None:
        with raises(SMTPServerDisconnected):
            send_email("no-reply@test.com", ["user@test.com"])

    def test_subject(self) -> None:
        with raises(SMTPServerDisconnected):
            send_email(
                "no-reply@test.com",
                ["user@test.com"],
                subject="Subject",
            )

    def test_contents_str(self) -> None:
        with raises(SMTPServerDisconnected):
            send_email(
                "no-reply@test.com",
                ["user@test.com"],
                subject="Subject",
                contents="contents",
            )

    def test_contents_airium(self) -> None:
        with yield_airium() as airium:
            _ = airium
        with raises(SMTPServerDisconnected):
            send_email(
                "no-reply@test.com",
                ["user@test.com"],
                subject="Subject",
                contents=airium,
            )

    def test_error(self) -> None:
        with raises(InvalidContentsError):
            send_email(
                "no-reply@test.com",
                ["user@test.com"],
                subject="Subject",
                contents=object(),
            )



from typing import Any

from beartype import beartype
from luigi import Parameter
from sqlalchemy.engine import Engine

from utilities.sqlalchemy import get_table_name


class EngineParameter(Parameter):
    """Parameter taking the value of a SQLAlchemy engine."""

    @beartype
    def normalize(self, engine: Engine, /) -> Engine:
        """Normalize an `Engine` argument."""
        return engine

    @beartype
    def serialize(self, engine: Engine, /) -> str:
        """serialize an `Engine` argument."""
        return engine.url.render_as_string()


class TableParameter(Parameter):
    """Parameter taking the value of a SQLAlchemy table."""

    @beartype
    def normalize(self, table_or_model: Any, /) -> Any:
        """Normalize a `Table` or model argument."""
        return table_or_model

    @beartype
    def serialize(self, table_or_model: Any, /) -> str:
        """serialize a `Table` or model argument."""
        return get_table_name(table_or_model)



from typing import NoReturn

from pytest import raises

from utilities.errors import NoUniqueArgError
from utilities.errors import redirect_error


class TestRedirectError:
    def test_generic_redirected_to_custom(self) -> None:
        with raises(self._CustomError):
            self._raises_custom("generic error")

    def test_generic_not_redirected_to_custom(self) -> None:
        with raises(ValueError, match="generic error"):
            self._raises_custom("something else")

    def _raises_generic(self) -> NoReturn:
        msg = "generic error"
        raise ValueError(msg)

    class _CustomError(ValueError):
        ...

    def _raises_custom(self, pattern: str, /) -> NoReturn:
        try:
            self._raises_generic()
        except ValueError as error:
            redirect_error(error, pattern, self._CustomError)

    def test_generic_with_no_unique_arg(self) -> None:
        def raises_generic() -> NoReturn:
            try:
                raise ValueError(0, 1)
            except ValueError as error:
                redirect_error(error, "error", RuntimeError)

        with raises(NoUniqueArgError):
            raises_generic()



from pathlib import Path

from click.testing import CliRunner

from utilities.rotate_logs import _key
from utilities.rotate_logs import _rotate_item
from utilities.rotate_logs import _rotate_items
from utilities.rotate_logs import _yield_items
from utilities.rotate_logs import main
from utilities.rotate_logs.classes import Item


class TestRotateLogs:
    def test_cli(self, tmp_path: Path) -> None:
        path = tmp_path.joinpath("file.log")
        path.touch()
        runner = CliRunner()
        args = ["--path", tmp_path.as_posix(), "--size", "0"]
        result = runner.invoke(main, args)
        assert result.exit_code == 0

    def test_dry_run(self, tmp_path: Path) -> None:
        path = tmp_path.joinpath("file.log")
        path.touch()
        runner = CliRunner()
        args = ["--path", tmp_path.as_posix(), "--size", "0", "--dry-run"]
        result = runner.invoke(main, args)
        assert result.exit_code == 0

    def test_rotate_items_one(self, tmp_path: Path) -> None:
        path = tmp_path.joinpath("file.log")
        path.touch()
        (items,) = _yield_items(path=tmp_path, size=0)
        _rotate_items(items, keep=1)
        contents = set(tmp_path.iterdir())
        expected = {tmp_path.joinpath("file.log.1")}
        assert contents == expected

    def test_rotate_items_two_growing(self, tmp_path: Path) -> None:
        path = tmp_path.joinpath("file.log")
        path.touch()
        tail = tmp_path.joinpath("file.log.1")
        tail.touch()
        (items,) = _yield_items(path=tmp_path, size=0)
        _rotate_items(items, keep=2)
        contents = set(tmp_path.iterdir())
        expected = {tail, tmp_path.joinpath("file.log.2")}
        assert contents == expected

    def test_rotate_items_two_at_limit(self, tmp_path: Path) -> None:
        path = tmp_path.joinpath("file.log")
        path.touch()
        tail = tmp_path.joinpath("file.log.1")
        tail.touch()
        (items,) = _yield_items(path=tmp_path, size=0)
        _rotate_items(items, keep=1)
        contents = set(tmp_path.iterdir())
        expected = {tail}
        assert contents == expected

    def test_rotate_item_moved(self, tmp_path: Path) -> None:
        path = tmp_path.joinpath("file.log")
        path.touch()
        ((item,),) = _yield_items(path=tmp_path, size=0)
        _rotate_item(item, keep=1)
        contents = set(tmp_path.iterdir())
        expected = {tmp_path.joinpath("file.log.1")}
        assert contents == expected

    def test_rotate_item_deleted(self, tmp_path: Path) -> None:
        head = tmp_path.joinpath("file.log")
        head.touch()
        tail = tmp_path.joinpath("file.log.1")
        tail.touch()
        (items,) = _yield_items(path=tmp_path, size=0)
        _, item = sorted(items, key=_key)
        _rotate_item(item, keep=1)
        contents = set(tmp_path.iterdir())
        expected = {head}
        assert contents == expected

    def test_yield_items_no_base(self, tmp_path: Path) -> None:
        items = frozenset(_yield_items(path=tmp_path, size=0))
        expected = frozenset()
        assert items == expected

    def test_yield_items_one_base_one_item(self, tmp_path: Path) -> None:
        head = tmp_path.joinpath("file.log")
        head.touch()
        items = frozenset(_yield_items(path=tmp_path, size=0))
        expected = frozenset([frozenset([Item(head, head)])])
        assert items == expected

    def test_yield_items_one_base_two_items(self, tmp_path: Path) -> None:
        head = tmp_path.joinpath("file.log")
        head.touch()
        tail = tmp_path.joinpath("file.log.1")
        tail.touch()
        items = frozenset(_yield_items(path=tmp_path, size=0))
        expected = frozenset(
            [frozenset([Item(head, head), Item(tail, head, num=1)])],
        )
        assert items == expected

    def test_yield_items_two_bases(self, tmp_path: Path) -> None:
        foo_head = tmp_path.joinpath("foo.log")
        foo_head.touch()
        foo_tail = tmp_path.joinpath("foo.log.1")
        foo_tail.touch()
        bar_head = tmp_path.joinpath("bar.log")
        bar_head.touch()
        items = frozenset(_yield_items(path=tmp_path, size=0))
        expected = frozenset(
            [
                frozenset(
                    [Item(foo_head, foo_head), Item(foo_tail, foo_head, num=1)],
                ),
                frozenset([Item(bar_head, bar_head)]),
            ],
        )
        assert items == expected



import datetime as dt
from contextlib import suppress
from enum import Enum as _Enum
from typing import Any
from typing import Generic
from typing import Optional
from typing import TypeVar

from beartype import beartype
from click import Context
from click import Parameter
from click import ParamType
from click import option

from utilities.datetime import ParseDateError
from utilities.datetime import ParseDateTimeError
from utilities.datetime import ParseTimeError
from utilities.datetime import TimedeltaError
from utilities.datetime import parse_date
from utilities.datetime import parse_datetime
from utilities.datetime import parse_time
from utilities.datetime import parse_timedelta
from utilities.logging import LogLevel


class Date(ParamType):
    """A date-valued parameter."""

    name = "date"

    @beartype
    def convert(
        self,
        value: Any,
        param: Optional[Parameter],
        ctx: Optional[Context],
    ) -> dt.date:
        """Convert a value into the `Date` type."""
        try:
            return parse_date(value)
        except ParseDateError:
            self.fail(f"Unable to parse {value}", param, ctx)


class DateTime(ParamType):
    """A datetime-valued parameter."""

    name = "datetime"

    @beartype
    def convert(
        self,
        value: Any,
        param: Optional[Parameter],
        ctx: Optional[Context],
    ) -> dt.date:
        """Convert a value into the `DateTime` type."""
        try:
            return parse_datetime(value)
        except ParseDateTimeError:
            self.fail(f"Unable to parse {value}", param, ctx)


class Time(ParamType):
    """A time-valued parameter."""

    name = "time"

    @beartype
    def convert(
        self,
        value: Any,
        param: Optional[Parameter],
        ctx: Optional[Context],
    ) -> dt.time:
        """Convert a value into the `Time` type."""
        try:
            return parse_time(value)
        except ParseTimeError:
            self.fail(f"Unable to parse {value}", param, ctx)


class Timedelta(ParamType):
    """A timedelta-valued parameter."""

    name = "timedelta"

    @beartype
    def convert(
        self,
        value: Any,
        param: Optional[Parameter],
        ctx: Optional[Context],
    ) -> dt.timedelta:
        """Convert a value into the `Timedelta` type."""
        try:
            return parse_timedelta(value)
        except TimedeltaError:
            self.fail(f"Unable to parse {value}", param, ctx)


_E = TypeVar("_E", bound=_Enum)


class Enum(ParamType, Generic[_E]):
    """An enum-valued parameter."""

    name = "enum"

    @beartype
    def __init__(self, enum: type[_E]) -> None:
        super().__init__()
        self._enum = enum

    @beartype
    def convert(
        self,
        value: Any,
        param: Optional[Parameter],
        ctx: Optional[Context],
    ) -> _E:
        """Convert a value into the `Enum` type."""
        els = {el for el in self._enum if el.name.lower() == value.lower()}
        with suppress(ValueError):
            (el,) = els
            return el
        return self.fail(f"Unable to parse {value}", param, ctx)


log_level_option = option(
    "--log-level",
    type=Enum(LogLevel),
    default=LogLevel.INFO,
    show_default=True,
    help="The logging level",
)



from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from smtplib import SMTP
from typing import Any
from typing import Optional

from beartype import beartype

from utilities.beartype import IterableStrs


@beartype
def send_email(
    from_: str,
    to: IterableStrs,
    /,
    *,
    subject: Optional[str] = None,
    contents: Any = None,
    subtype: str = "plain",
    host: str = "",
    port: int = 0,
) -> None:
    """Send an email."""
    message = MIMEMultipart()
    message["From"] = from_
    message["To"] = ",".join(to)
    if subject is not None:
        message["Subject"] = subject
    if contents is not None:
        if isinstance(contents, str):
            text = MIMEText(contents, subtype)
        else:
            try:
                from airium import Airium
            except ModuleNotFoundError:  # pragma: no cover
                raise InvalidContentsError(contents) from None
            else:
                if not isinstance(contents, Airium):
                    raise InvalidContentsError(contents)
                text = MIMEText(str(contents), "html")
        message.attach(text)
    with SMTP(host=host, port=port) as smtp:
        _ = smtp.send_message(message)


class InvalidContentsError(TypeError):
    """Raised when an invalid set of contents is encountered."""



import datetime as dt
from collections.abc import Iterator
from contextlib import contextmanager
from pathlib import Path
from zoneinfo import ZoneInfo

from beartype import beartype
from pyinstrument.profiler import Profiler

from utilities.pathlib import PathLike

_CWD = Path.cwd()


@contextmanager
@beartype
def profile(*, path: PathLike = _CWD) -> Iterator[None]:
    """Profile the contents of a block."""
    with Profiler() as profiler:
        yield
    now = dt.datetime.now(tz=ZoneInfo("UTC"))
    filename = Path(path, f"profile__{now:%Y%m%dT%H%M%S}.html")
    with open(filename, mode="w") as fh:
        _ = fh.write(profiler.output_html())



from typing import Optional

from hypothesis import given
from hypothesis.errors import InvalidArgument
from hypothesis.extra.numpy import array_shapes
from hypothesis.strategies import DataObject
from hypothesis.strategies import booleans
from hypothesis.strategies import data
from hypothesis.strategies import floats
from hypothesis.strategies import none
from numpy import iinfo
from numpy import int64
from numpy import isinf
from numpy import isnan

from utilities.hypothesis import assume_does_not_raise
from utilities.hypothesis.numpy import bool_arrays
from utilities.hypothesis.numpy import float_arrays
from utilities.hypothesis.numpy import int64s
from utilities.hypothesis.numpy import int_arrays
from utilities.hypothesis.typing import Shape


class TestBoolArrays:
    @given(data=data(), shape=array_shapes())
    def test_main(self, data: DataObject, shape: Shape) -> None:
        array = data.draw(bool_arrays(shape=shape))
        assert array.dtype == bool
        assert array.shape == shape


class TestFloatArrays:
    @given(
        data=data(),
        shape=array_shapes(),
        min_value=floats() | none(),
        max_value=floats() | none(),
        allow_nan=booleans() | none(),
        allow_infinity=booleans() | none(),
    )
    def test_main(
        self,
        data: DataObject,
        shape: Shape,
        min_value: Optional[float],
        max_value: Optional[float],
        allow_nan: Optional[bool],
        allow_infinity: Optional[bool],
    ) -> None:
        with assume_does_not_raise(InvalidArgument):
            array = data.draw(
                float_arrays(
                    shape=shape,
                    min_value=min_value,
                    max_value=max_value,
                    allow_nan=allow_nan,
                    allow_infinity=allow_infinity,
                ),
            )
        assert array.dtype == float
        assert array.shape == shape
        if allow_nan is False:
            assert (~isnan(array)).all()
        if allow_infinity is False:
            assert (~isinf(array)).all()


class TestIntArrays:
    @given(
        data=data(),
        shape=array_shapes(),
        min_value=int64s() | none(),
        max_value=int64s() | none(),
    )
    def test_main(
        self,
        data: DataObject,
        shape: Shape,
        min_value: Optional[int],
        max_value: Optional[int],
    ) -> None:
        with assume_does_not_raise(InvalidArgument):
            array = data.draw(
                int_arrays(
                    shape=shape,
                    min_value=min_value,
                    max_value=max_value,
                ),
            )
        assert array.dtype == int
        assert array.shape == shape


class TestInt64s:
    @given(x=int64s())
    def test_main(self, x: int) -> None:
        assert isinstance(x, int)
        info = iinfo(int64)
        assert info.min <= x <= info.max



import datetime as dt
from typing import Any

from beartype import beartype
from hypothesis.strategies import SearchStrategy
from hypothesis.strategies import composite
from hypothesis.strategies import dates
from hypothesis.strategies import datetimes

from utilities.datetime import UTC
from utilities.hypothesis import lift_draw
from utilities.pandas import TIMESTAMP_MAX_AS_DATE
from utilities.pandas import TIMESTAMP_MAX_AS_DATETIME
from utilities.pandas import TIMESTAMP_MIN_AS_DATE
from utilities.pandas import TIMESTAMP_MIN_AS_DATETIME


@beartype
def dates_pd(
    *,
    min_value: dt.date = TIMESTAMP_MIN_AS_DATE,
    max_value: dt.date = TIMESTAMP_MAX_AS_DATE,
) -> SearchStrategy[dt.date]:
    """Strategy for generating dates which can become Timestamps."""
    return dates(min_value=min_value, max_value=max_value)


@composite
@beartype
def datetimes_pd(
    _draw: Any,
    /,
    *,
    min_value: dt.datetime = TIMESTAMP_MIN_AS_DATETIME,
    max_value: dt.datetime = TIMESTAMP_MAX_AS_DATETIME,
) -> dt.datetime:
    """Strategy for generating datetimes which can become Timestamps."""
    draw = lift_draw(_draw)
    datetime = draw(
        datetimes(
            min_value=min_value.replace(tzinfo=None),
            max_value=max_value.replace(tzinfo=None),
        ),
    )
    return datetime.replace(tzinfo=UTC)



from collections.abc import Iterator
from collections.abc import Mapping
from functools import partial
from itertools import chain
from itertools import repeat
from itertools import starmap
from pathlib import Path
from subprocess import PIPE
from subprocess import CalledProcessError
from subprocess import check_output
from typing import Any
from typing import Optional

from beartype import beartype

from utilities.os import temp_environ
from utilities.pathlib import PathLike

_CWD = Path.cwd()


@beartype
def get_shell_output(
    cmd: str,
    /,
    *,
    cwd: PathLike = _CWD,
    activate: Optional[PathLike] = None,
    env: Optional[Mapping[str, Optional[str]]] = None,
) -> str:
    """Get the output of a shell call.

    Optionally, activate a virtual environment if necessary.
    """
    cwd = Path(cwd)
    if activate is not None:
        activates = list(cwd.rglob("activate"))
        if (n := len(activates)) == 0:
            raise NoActivateError(cwd)
        if n == 1:
            cmd = f"source {activates[0]}; {cmd}"
        else:
            raise MultipleActivateError(activates)
    with temp_environ(env):
        return check_output(cmd, stderr=PIPE, shell=True, cwd=cwd, text=True)


class NoActivateError(ValueError):
    """Raised when no `activate` script can be found."""


class MultipleActivateError(ValueError):
    """Raised when multiple `activate` scripts can be found."""


@beartype
def tabulate_called_process_error(error: CalledProcessError, /) -> str:
    """Tabulate the components of a CalledProcessError."""
    mapping = {
        "cmd": error.cmd,
        "returncode": error.returncode,
        "stdout": error.stdout,
        "stderr": error.stderr,
    }
    max_key_len = max(map(len, mapping))
    tabulate = partial(_tabulate, buffer=max_key_len + 1)
    return "\n".join(starmap(tabulate, mapping.items()))


@beartype
def _tabulate(key: str, value: Any, /, *, buffer: int) -> str:
    template = f"{{:{buffer}}}{{}}"

    @beartype
    def yield_lines() -> Iterator[str]:
        keys = chain([key], repeat(buffer * " "))
        value_lines = str(value).splitlines()
        for k, v in zip(keys, value_lines):
            yield template.format(k, v)

    return "\n".join(yield_lines())



import datetime as dt
from typing import Any
from typing import cast

from beartype import beartype
from pandas import NaT
from pandas import Timestamp

from utilities.datetime import UTC

Int64 = "Int64"
boolean = "boolean"
string = "string"


@beartype
def timestamp_to_date(timestamp: Any, /) -> dt.date:
    """Convert a timestamp to a date."""
    return timestamp_to_datetime(timestamp).date()


@beartype
def timestamp_to_datetime(timestamp: Any, /) -> dt.datetime:
    """Convert a timestamp to a datetime."""
    if timestamp is NaT:
        msg = f"Invalid value: {timestamp}"
        raise ValueError(msg)
    datetime = cast(dt.datetime, timestamp.to_pydatetime())
    if datetime.tzinfo is None:
        return datetime.replace(tzinfo=UTC)
    return datetime


@beartype
def _timestamp_minmax_to_date(
    timestamp: Timestamp,
    method_name: str,
    /,
) -> dt.date:
    """Get the maximum Timestamp as a date."""
    method = getattr(timestamp, method_name)
    rounded = cast(Timestamp, method("D"))
    return timestamp_to_date(rounded)


TIMESTAMP_MIN_AS_DATE = _timestamp_minmax_to_date(Timestamp.min, "ceil")
TIMESTAMP_MAX_AS_DATE = _timestamp_minmax_to_date(Timestamp.max, "floor")


@beartype
def _timestamp_minmax_to_datetime(
    timestamp: Timestamp,
    method_name: str,
    /,
) -> dt.datetime:
    """Get the maximum Timestamp as a datetime."""
    method = getattr(timestamp, method_name)
    rounded = cast(Timestamp, method("us"))
    return timestamp_to_datetime(rounded)


TIMESTAMP_MIN_AS_DATETIME = _timestamp_minmax_to_datetime(Timestamp.min, "ceil")
TIMESTAMP_MAX_AS_DATETIME = _timestamp_minmax_to_datetime(
    Timestamp.max,
    "floor",
)



from pathlib import Path
from subprocess import CalledProcessError
from subprocess import check_call

from pytest import raises

from utilities.subprocess import MultipleActivateError
from utilities.subprocess import NoActivateError
from utilities.subprocess import get_shell_output
from utilities.subprocess import tabulate_called_process_error
from utilities.text import strip_and_dedent


class TestGetShellOutput:
    def test_main(self) -> None:
        output = get_shell_output("ls")
        assert any(line == "pyproject.toml" for line in output.splitlines())

    def test_activate(self, tmp_path: Path) -> None:
        venv = tmp_path.joinpath(".venv")
        activate = venv.joinpath("activate")
        activate.parent.mkdir(parents=True)
        activate.touch()
        _ = get_shell_output("ls", cwd=venv, activate=venv)

    def test_no_activate(self, tmp_path: Path) -> None:
        venv = tmp_path.joinpath(".venv")
        with raises(NoActivateError):
            _ = get_shell_output("ls", cwd=venv, activate=venv)

    def test_multiple_activates(self, tmp_path: Path) -> None:
        venv = tmp_path.joinpath(".venv")
        for i in range(2):
            activate = venv.joinpath(str(i), "activate")
            activate.parent.mkdir(parents=True)
            activate.touch()
        with raises(MultipleActivateError):
            _ = get_shell_output("ls", cwd=venv, activate=venv)


class TestTabulateCalledProcessError:
    def test_main(self) -> None:
        def which() -> None:
            _ = check_call(["which"], text=True)

        try:
            which()
        except CalledProcessError as error:
            result = tabulate_called_process_error(error)
            expected = """
                cmd        ['which']
                returncode 1
                stdout     None
                stderr     None
            """
            assert result == strip_and_dedent(expected)
        else:
            with raises(CalledProcessError):
                which()



import builtins
from collections.abc import Iterable
from collections.abc import Iterator
from contextlib import contextmanager
from os import getenv
from pathlib import Path
from re import search
from string import ascii_letters
from string import printable
from typing import Any
from typing import Optional
from typing import Protocol
from typing import TypeVar
from typing import cast
from typing import overload

from beartype import beartype
from hypothesis import Verbosity
from hypothesis import assume
from hypothesis import settings
from hypothesis.strategies import DrawFn
from hypothesis.strategies import SearchStrategy
from hypothesis.strategies import characters
from hypothesis.strategies import composite
from hypothesis.strategies import lists
from hypothesis.strategies import text
from hypothesis.strategies import uuids

from utilities.hypothesis.typing import MaybeSearchStrategy
from utilities.tempfile import TemporaryDirectory
from utilities.tempfile import gettempdir
from utilities.text import ensure_str


@contextmanager
@beartype
def assume_does_not_raise(
    *exceptions: type[Exception],
    match: Optional[str] = None,
) -> Iterator[None]:
    """Assume a set of exceptions are not raised.

    Optionally filter on the string representation of the exception.
    """
    try:
        yield
    except exceptions as caught:
        if match is None:
            _ = assume(condition=False)
        else:
            (msg,) = caught.args
            if search(match, ensure_str(msg)):
                _ = assume(condition=False)
            else:
                raise


_MDF = TypeVar("_MDF")


class _MaybeDrawFn(Protocol):
    @overload
    def __call__(self, obj: SearchStrategy[_MDF], /) -> _MDF:
        ...

    @overload
    def __call__(self, obj: MaybeSearchStrategy[_MDF], /) -> _MDF:
        ...

    def __call__(self, obj: MaybeSearchStrategy[_MDF], /) -> _MDF:
        raise NotImplementedError(obj)  # pragma: no cover


def lift_draw(draw: DrawFn, /) -> _MaybeDrawFn:
    """Lift the `draw` function to handle non-`SearchStrategy` types."""

    @beartype
    def func(obj: MaybeSearchStrategy[_MDF], /) -> _MDF:
        return draw(obj) if isinstance(obj, SearchStrategy) else obj

    return func


_TLFL = TypeVar("_TLFL")


@composite
@beartype
def lists_fixed_length(
    _draw: Any,
    strategy: SearchStrategy[_TLFL],
    size: MaybeSearchStrategy[int],
    /,
    *,
    unique: MaybeSearchStrategy[bool] = False,
    sorted: MaybeSearchStrategy[bool] = False,  # noqa: A002
) -> list[_TLFL]:
    """Strategy for generating lists of a fixed length."""
    draw = lift_draw(_draw)
    size_ = draw(size)
    elements = draw(
        lists(strategy, min_size=size_, max_size=size_, unique=draw(unique)),
    )
    if draw(sorted):
        return builtins.sorted(cast(Iterable[Any], elements))
    return elements


@beartype
def setup_hypothesis_profiles() -> None:
    """Set up the hypothesis profiles."""
    kwargs = {
        "deadline": None,
        "print_blob": True,
        "report_multiple_bugs": False,
    }
    settings.register_profile("default", max_examples=100, **kwargs)
    settings.register_profile("dev", max_examples=10, **kwargs)
    settings.register_profile("ci", max_examples=1000, **kwargs)
    settings.register_profile(
        "debug",
        max_examples=10,
        verbosity=Verbosity.verbose,
        **kwargs,
    )
    settings.load_profile(getenv("HYPOTHESIS_PROFILE", "default"))


@composite
@beartype
def temp_dirs(_draw: Any, /) -> TemporaryDirectory:
    """Search strategy for temporary directories."""
    dir_ = gettempdir().joinpath("hypothesis")
    dir_.mkdir(exist_ok=True)
    uuid = _draw(uuids())
    return TemporaryDirectory(prefix=f"{uuid}__", dir=dir_.as_posix())


@composite
@beartype
def temp_paths(_draw: Any, /) -> Path:
    """Search strategy for paths to temporary directories."""
    temp_dir = _draw(temp_dirs())
    root = temp_dir.name
    cls = type(root)

    class SubPath(cls):
        _temp_dir = temp_dir

    return SubPath(root)


@beartype
def text_ascii(
    *,
    min_size: MaybeSearchStrategy[int] = 0,
    max_size: MaybeSearchStrategy[Optional[int]] = None,
) -> SearchStrategy[str]:
    """Strategy for generating ASCII text."""
    return _draw_text(
        characters(whitelist_categories=[], whitelist_characters=ascii_letters),
        min_size=min_size,
        max_size=max_size,
    )


@beartype
def text_clean(
    *,
    min_size: MaybeSearchStrategy[int] = 0,
    max_size: MaybeSearchStrategy[Optional[int]] = None,
) -> SearchStrategy[str]:
    """Strategy for generating clean text."""
    return _draw_text(
        characters(blacklist_categories=["Z", "C"]),
        min_size=min_size,
        max_size=max_size,
    )


@beartype
def text_printable(
    *,
    min_size: MaybeSearchStrategy[int] = 0,
    max_size: MaybeSearchStrategy[Optional[int]] = None,
) -> SearchStrategy[str]:
    """Strategy for generating printable text."""
    return _draw_text(
        characters(whitelist_categories=[], whitelist_characters=printable),
        min_size=min_size,
        max_size=max_size,
    )


@composite
def _draw_text(
    _draw: DrawFn,
    alphabet: MaybeSearchStrategy[str],
    /,
    *,
    min_size: MaybeSearchStrategy[int] = 0,
    max_size: MaybeSearchStrategy[Optional[int]] = None,
) -> str:
    draw = lift_draw(_draw)
    return draw(
        text(alphabet, min_size=draw(min_size), max_size=draw(max_size)),
    )



import datetime as dt
from typing import cast

from hypothesis import given
from pandas import Timestamp
from pandas import to_datetime

from utilities.hypothesis.pandas import dates_pd
from utilities.hypothesis.pandas import datetimes_pd


class TestDatesPd:
    @given(date=dates_pd())
    def test_main(self, date: dt.date) -> None:
        assert isinstance(to_datetime(cast(Timestamp, date)), dt.date)


class TestDatetimesPd:
    @given(date=datetimes_pd())
    def test_main(self, date: dt.datetime) -> None:
        assert isinstance(to_datetime(cast(Timestamp, date)), dt.datetime)



import datetime as dt
from collections.abc import Callable
from enum import Enum as _Enum
from enum import auto
from typing import Any

from click import ParamType
from click import argument
from click import command
from click import echo
from click.testing import CliRunner
from hypothesis import given
from hypothesis.strategies import DataObject
from hypothesis.strategies import SearchStrategy
from hypothesis.strategies import data
from hypothesis.strategies import dates
from hypothesis.strategies import datetimes
from hypothesis.strategies import just
from hypothesis.strategies import sampled_from
from hypothesis.strategies import timedeltas
from hypothesis.strategies import times
from pytest import mark
from pytest import param

from utilities.click import Date
from utilities.click import DateTime
from utilities.click import Enum
from utilities.click import Time
from utilities.click import Timedelta
from utilities.click import log_level_option
from utilities.datetime import UTC
from utilities.datetime import serialize_date
from utilities.datetime import serialize_datetime
from utilities.datetime import serialize_time
from utilities.datetime import serialize_timedelta
from utilities.logging import LogLevel


def runners() -> SearchStrategy[CliRunner]:
    return just(CliRunner())


class TestParameters:
    @given(data=data())
    @mark.parametrize(
        ("param", "cls", "strategy", "serialize"),
        [
            param(Date(), dt.date, dates(), serialize_date),
            param(
                DateTime(),
                dt.datetime,
                datetimes(timezones=just(UTC)),
                serialize_datetime,
            ),
            param(Time(), dt.time, times(), serialize_time),
            param(Timedelta(), dt.timedelta, timedeltas(), serialize_timedelta),
        ],
    )
    def test_success(
        self,
        data: DataObject,
        param: ParamType,
        cls: Any,
        strategy: SearchStrategy[Any],
        serialize: Callable[[Any], str],
    ) -> None:
        runner = CliRunner()

        @command()
        @argument("value", type=param)
        def cli(*, value: cls) -> None:
            echo(f"value = {serialize(value)}")

        value_str = serialize(data.draw(strategy))
        result = runner.invoke(cli, [value_str])
        assert result.exit_code == 0
        assert result.stdout == f"value = {value_str}\n"

        result = runner.invoke(cli, ["error"])
        assert result.exit_code == 2


class Truth(_Enum):
    true = auto()
    false = auto()


@command()
@argument("truth", type=Enum(Truth))
def uses_enum(*, truth: Truth) -> None:
    echo(f"truth = {truth}")


class TestEnum:
    @given(data=data(), runner=runners(), truth=sampled_from(Truth))
    def test_success(
        self,
        data: DataObject,
        runner: CliRunner,
        truth: Truth,
    ) -> None:
        name = truth.name
        as_str = data.draw(sampled_from([name, name.lower()]))
        result = runner.invoke(uses_enum, [as_str])
        assert result.exit_code == 0
        assert result.stdout == f"truth = {truth}\n"

    @given(runner=runners())
    def test_failure(self, runner: CliRunner) -> None:
        result = runner.invoke(uses_enum, ["not_an_element"])
        assert result.exit_code == 2


@command()
@log_level_option
def uses_log_level(*, log_level: LogLevel) -> None:
    echo(f"log_level = {log_level}")


class TestLogLevelOption:
    @given(runner=runners(), log_level=sampled_from(LogLevel))
    def test_main(self, runner: CliRunner, log_level: LogLevel) -> None:
        result = runner.invoke(uses_log_level, [f"--log-level={log_level}"])
        assert result.exit_code == 0
        assert result.stdout == f"log_level = {log_level}\n"



from collections.abc import Iterable
from collections.abc import Iterator
from collections.abc import Mapping
from contextlib import contextmanager
from contextlib import suppress
from os import environ
from os import getenv
from typing import Optional
from typing import cast

from beartype import beartype


@contextmanager
@beartype
def temp_environ(
    env: Optional[Mapping[str, Optional[str]]] = None,
    **env_kwargs: Optional[str],
) -> Iterator[None]:
    """Context manager with temporary environment variable set."""
    all_env = (
        cast(dict[str, Optional[str]], {}) if env is None else env
    ) | env_kwargs
    prev = list(zip(all_env, map(getenv, all_env)))
    _apply_environment(all_env.items())
    try:
        yield
    finally:
        _apply_environment(prev)


@beartype
def _apply_environment(items: Iterable[tuple[str, Optional[str]]], /) -> None:
    for key, value in items:
        if value is None:
            with suppress(KeyError):
                del environ[key]
        else:
            environ[key] = value



from typing import TypeVar
from typing import Union

from hypothesis.strategies import SearchStrategy

_T = TypeVar("_T")
MaybeSearchStrategy = Union[_T, SearchStrategy[_T]]


Shape = Union[int, tuple[int, ...]]



from collections.abc import Callable
from collections.abc import Iterable
from collections.abc import Iterator
from contextlib import contextmanager
from functools import reduce
from operator import ge
from operator import le
from re import search
from typing import Any
from typing import Literal
from typing import Optional
from typing import Union
from typing import cast

from beartype import beartype
from more_itertools import chunked
from sqlalchemy import Table
from sqlalchemy import and_
from sqlalchemy import case
from sqlalchemy import create_engine as _create_engine
from sqlalchemy.dialects.mssql import dialect as mssql_dialect
from sqlalchemy.dialects.mysql import dialect as mysql_dialect
from sqlalchemy.dialects.oracle import dialect as oracle_dialect
from sqlalchemy.dialects.postgresql import dialect as postgresql_dialect
from sqlalchemy.dialects.sqlite import dialect as sqlite_dialect
from sqlalchemy.engine import URL
from sqlalchemy.engine import Connection
from sqlalchemy.engine import Engine
from sqlalchemy.exc import DatabaseError
from sqlalchemy.exc import OperationalError
from sqlalchemy.pool import NullPool
from sqlalchemy.pool import Pool
from sqlalchemy.sql import Selectable

from utilities.typing import never


@beartype
def columnwise_max(*columns: Any) -> Any:
    """Compute the columnwise max of a number of columns."""
    return _columnwise_minmax(*columns, op=ge)


@beartype
def columnwise_min(*columns: Any) -> Any:
    """Compute the columnwise min of a number of columns."""
    return _columnwise_minmax(*columns, op=le)


@beartype
def _columnwise_minmax(*columns: Any, op: Callable[[Any, Any], Any]) -> Any:
    """Compute the columnwise min of a number of columns."""

    @beartype
    def func(x: Any, y: Any, /) -> Any:
        x_none = x.is_(None)
        y_none = y.is_(None)
        col = case(
            (and_(x_none, y_none), None),
            (and_(~x_none, y_none), x),
            (and_(x_none, ~y_none), y),
            (op(x, y), x),
            else_=y,
        )
        # try auto-label
        names = {
            value
            for col in [x, y]
            if (value := getattr(col, "name", None)) is not None
        }
        try:
            (name,) = names
        except ValueError:
            return col
        else:
            return col.label(name)

    return reduce(func, columns)


@beartype
def create_engine(
    drivername: str,
    /,
    *,
    username: Optional[str] = None,
    password: Optional[str] = None,
    host: Optional[str] = None,
    port: Optional[int] = None,
    database: Optional[str] = None,
    poolclass: Optional[type[Pool]] = NullPool,
) -> Engine:
    """Create a SQLAlchemy engine."""
    url = URL.create(
        drivername,
        username=username,
        password=password,
        host=host,
        port=port,
        database=database,
    )
    return _create_engine(url, future=True, poolclass=poolclass)


Dialect = Literal["mssql", "mysql", "oracle", "postgresql", "sqlite"]


@beartype
def get_dialect(engine_or_conn: Union[Engine, Connection], /) -> Dialect:
    """Get the dialect of a database."""
    if isinstance(dialect := engine_or_conn.dialect, mssql_dialect):
        return "mssql"
    if isinstance(dialect, mysql_dialect):
        return "mysql"
    if isinstance(dialect, oracle_dialect):
        return "oracle"
    if isinstance(dialect, postgresql_dialect):
        return "postgresql"
    if isinstance(dialect, sqlite_dialect):
        return "sqlite"
    msg = f"{dialect=}"  # pragma: no cover
    raise UnsupportedDialectError(msg)  # pragma: no cover


class UnsupportedDialectError(TypeError):
    """Raised when a dialect is unsupported."""


@beartype
def ensure_table_created(table_or_model: Any, engine: Engine, /) -> None:
    """Ensure a table is created."""
    table = get_table(table_or_model)
    try:
        with engine.begin() as conn:
            table.create(conn)
    # note that OperationalError < DatabaseError
    except OperationalError as error:
        # sqlite
        (msg,) = error.args
        if not search("table .* already exists", msg):  # pragma: no cover
            raise
    except DatabaseError as error:  # pragma: no cover
        # oracle
        (msg,) = error.args
        if not search(
            "ORA-00955: name is already used by an existing object",
            msg,
        ):
            raise


@beartype
def ensure_table_dropped(table_or_model: Any, engine: Engine, /) -> None:
    """Ensure a table is dropped."""
    table = get_table(table_or_model)
    try:
        with engine.begin() as conn:
            table.drop(conn)
    # note that OperationalError < DatabaseError
    except OperationalError as error:
        # sqlite
        (msg,) = error.args
        if not search("no such table", msg):  # pragma: no cover
            raise
    except DatabaseError as error:  # pragma: no cover
        # oracle
        (msg,) = error.args
        if not search("ORA-00942: table or view does not exist", msg):
            raise


@beartype
def get_column_names(table_or_model: Any, /) -> list[str]:
    """Get the column names from a table or model."""
    return [col.name for col in get_columns(table_or_model)]


@beartype
def get_columns(table_or_model: Any, /) -> list[Any]:
    """Get the columns from a table or model."""
    return list(get_table(table_or_model).columns)


@beartype
def get_table(table_or_model: Any, /) -> Table:
    """Get the table from a ORM model."""
    if isinstance(table_or_model, Table):
        return table_or_model
    return table_or_model.__table__


@beartype
def get_table_name(table_or_model: Any, /) -> str:
    """Get the table name from a ORM model."""
    return get_table(table_or_model).name


@contextmanager
@beartype
def yield_connection(
    engine_or_conn: Union[Engine, Connection],
    /,
) -> Iterator[Connection]:
    """Yield a connection."""
    if isinstance(engine_or_conn, Engine):
        with engine_or_conn.begin() as conn:
            yield conn
    else:
        yield engine_or_conn


@beartype
def yield_in_clause_rows(
    sel: Selectable,
    column: Any,
    values: Iterable[Any],
    engine_or_conn: Union[Engine, Connection],
    /,
    *,
    chunk_size: Optional[int] = None,
    frac: float = 0.95,
) -> Iterator[Any]:
    """Yield the rows from an `in` clause."""
    if chunk_size is None:
        dialect = get_dialect(engine_or_conn)
        if dialect == "mssql":  # pragma: no cover
            max_params = 2100
        elif dialect == "mysql":  # pragma: no cover
            max_params = 65535
        elif dialect == "oracle":  # pragma: no cover
            max_params = 1000
        elif dialect == "postgresql":  # pragma: no cover
            max_params = 32767
        elif dialect == "sqlite":
            max_params = 100
        else:
            return never(dialect)  # pragma: no cover
        chunk_size_use = round(frac * max_params)
    else:
        chunk_size_use = chunk_size
    with yield_connection(engine_or_conn) as conn:
        for values_i in chunked(values, chunk_size_use):
            sel_i: Selectable = cast(Any, sel).where(column.in_(values_i))
            yield from conn.execute(sel_i).all()
    return None



from collections.abc import Callable
from functools import partial
from operator import le
from operator import lt
from types import ModuleType
from typing import Any
from typing import Optional
from typing import Union

from pytest import mark
from pytest import param

from tests.modules import package_with
from tests.modules import package_without
from tests.modules import standalone
from utilities.modules import yield_module_contents
from utilities.modules import yield_modules


class TestYieldModules:
    @mark.parametrize(
        ("module", "recursive", "expected"),
        [
            param(standalone, False, 1),
            param(standalone, True, 1),
            param(package_without, False, 2),
            param(package_without, True, 2),
            param(package_with, False, 2),
            param(package_with, True, 5),
        ],
    )
    def test_main(
        self,
        module: ModuleType,
        recursive: bool,
        expected: int,
    ) -> None:
        assert len(list(yield_modules(module, recursive=recursive))) == expected


class TestYieldModuleContents:
    @mark.parametrize(
        ("type_", "predicate", "expected"),
        [
            param(None, None, 14),
            param(int, None, 3),
            param(float, None, 3),
            param((int, float), None, 6),
            param(int, partial(le, 0), 2),
            param(int, partial(lt, 0), 1),
            param(float, partial(le, 0), 2),
            param(float, partial(lt, 0), 1),
        ],
    )
    def test_main(
        self,
        type_: Optional[Union[type, tuple[type, ...]]],
        predicate: Callable[[Any], bool],
        expected: int,
    ) -> None:
        it = yield_module_contents(standalone, type=type_, predicate=predicate)
        assert len(list(it)) == expected



from functools import partial
from pathlib import Path
from typing import cast

from hypothesis import given
from hypothesis import settings
from hypothesis.strategies import booleans
from luigi import BoolParameter
from luigi import Task
from luigi.notifications import smtp

from utilities.hypothesis.luigi import task_namespaces
from utilities.luigi import PathTarget
from utilities.luigi import _yield_task_classes
from utilities.luigi import build
from utilities.luigi import clone
from utilities.luigi import get_dependencies_downstream
from utilities.luigi import get_dependencies_upstream
from utilities.luigi import get_task_classes


class TestBuild:
    @given(namespace=task_namespaces())
    def test_main(self, namespace: str) -> None:
        class Example(Task):
            task_namespace = namespace

        _ = build([Example()], local_scheduler=True)


class TestClone:
    @given(namespace=task_namespaces(), truth=booleans())
    def test_main(self, namespace: str, truth: bool) -> None:
        class A(Task):
            task_namespace = namespace

            truth = cast(bool, BoolParameter())

        class B(Task):
            task_namespace = namespace

            truth = cast(bool, BoolParameter())

        a = A(truth)
        result = clone(a, B)
        expected = B(truth)
        assert result is expected


class TestPathTarget:
    def test_main(self, tmp_path: Path) -> None:
        target = PathTarget(path := tmp_path.joinpath("file"))
        assert isinstance(target.path, Path)
        assert not target.exists()
        path.touch()
        assert target.exists()


class TestGetDependencies:
    @given(namespace=task_namespaces())
    @settings(max_examples=1)
    def test_main(self, namespace: str) -> None:
        class A(Task):
            task_namespace = namespace

        class B(Task):
            task_namespace = namespace

            def requires(self) -> A:
                return clone(self, A)

        class C(Task):
            task_namespace = namespace

            def requires(self) -> B:
                return clone(self, B)

        a, b, c = A(), B(), C()
        ((up_a, down_a), (up_b, down_b), (up_c, down_c)) = map(
            self._get_sets,
            [a, b, c],
        )
        assert up_a == set()
        assert down_a == {b}
        assert up_b == {a}
        assert down_b == {c}
        assert up_c == {b}
        assert down_c == set()

        (
            (up_a_rec, down_a_rec),
            (up_b_rec, down_b_rec),
            (up_c_rec, down_c_rec),
        ) = map(partial(self._get_sets, recursive=True), [a, b, c])
        assert up_a_rec == set()
        assert down_a_rec == {b, c}
        assert up_b_rec == {a}
        assert down_b_rec == {c}
        assert up_c_rec == {a, b}
        assert down_c_rec == set()

    @staticmethod
    def _get_sets(
        task: Task,
        /,
        *,
        recursive: bool = False,
    ) -> tuple[set[Task], set[Task]]:
        return set(get_dependencies_upstream(task, recursive=recursive)), set(
            get_dependencies_downstream(task, recursive=recursive),
        )


class TestGetTaskClasses:
    @given(namespace=task_namespaces())
    @settings(max_examples=1)
    def test_main(self, namespace: str) -> None:
        class Example(Task):
            task_namespace = namespace

        assert Example in get_task_classes()

    def test_notifications(self) -> None:
        assert smtp not in _yield_task_classes()

    @given(namespace=task_namespaces())
    @settings(max_examples=1)
    def test_filter(self, namespace: str) -> None:
        class Parent(Task):
            task_namespace = namespace

        class Child(Parent):
            ...

        result = get_task_classes(cls=Parent)
        expected = frozenset([Child])
        assert result == expected



from typing import Any
from typing import Optional

from beartype import beartype
from hypothesis.extra.numpy import array_shapes
from hypothesis.extra.numpy import arrays
from hypothesis.extra.numpy import from_dtype
from hypothesis.strategies import SearchStrategy
from hypothesis.strategies import booleans
from hypothesis.strategies import composite
from hypothesis.strategies import floats
from hypothesis.strategies import integers
from hypothesis.strategies import nothing
from numpy import bool_
from numpy import dtype
from numpy import float64
from numpy import iinfo
from numpy import int64
from numpy.typing import NDArray

from utilities.hypothesis import lift_draw
from utilities.hypothesis.typing import MaybeSearchStrategy
from utilities.hypothesis.typing import Shape

_ARRAY_SHAPES = array_shapes()


@composite
@beartype
def bool_arrays(
    _draw: Any,
    /,
    *,
    shape: MaybeSearchStrategy[Shape] = _ARRAY_SHAPES,
    unique: MaybeSearchStrategy[bool] = False,
) -> NDArray[bool_]:
    """Strategy for generating arrays of booleans."""
    draw = lift_draw(_draw)
    return draw(
        arrays(
            bool,
            draw(shape),
            elements=booleans(),
            fill=nothing(),
            unique=draw(unique),
        ),
    )


@composite
@beartype
def float_arrays(
    _draw: Any,
    /,
    *,
    shape: MaybeSearchStrategy[Shape] = _ARRAY_SHAPES,
    min_value: MaybeSearchStrategy[Optional[float]] = None,
    max_value: MaybeSearchStrategy[Optional[float]] = None,
    allow_nan: MaybeSearchStrategy[Optional[bool]] = None,
    allow_infinity: MaybeSearchStrategy[Optional[bool]] = None,
    unique: MaybeSearchStrategy[bool] = False,
) -> NDArray[float64]:
    """Strategy for generating arrays of floats."""
    draw = lift_draw(_draw)
    elements = floats(
        min_value=draw(min_value),
        max_value=draw(max_value),
        allow_nan=draw(allow_nan),
        allow_infinity=draw(allow_infinity),
    )
    return draw(
        arrays(
            float,
            draw(shape),
            elements=elements,
            fill=nothing(),
            unique=draw(unique),
        ),
    )


@composite
@beartype
def int_arrays(
    _draw: Any,
    /,
    *,
    shape: MaybeSearchStrategy[Shape] = _ARRAY_SHAPES,
    min_value: MaybeSearchStrategy[Optional[int]] = None,
    max_value: MaybeSearchStrategy[Optional[int]] = None,
    unique: MaybeSearchStrategy[bool] = False,
) -> NDArray[int64]:
    """Strategy for generating arrays of ints."""
    draw = lift_draw(_draw)
    info = iinfo(int64)
    min_value_, max_value_ = draw(min_value), draw(max_value)
    min_value_use = info.min if min_value_ is None else min_value_
    max_value_use = info.max if max_value_ is None else max_value_
    elements = integers(min_value=min_value_use, max_value=max_value_use)
    return draw(
        arrays(
            int,
            draw(shape),
            elements=elements,
            fill=nothing(),
            unique=draw(unique),
        ),
    )


@beartype
def int64s() -> SearchStrategy[int]:
    """Strategy for generating int64s."""
    return from_dtype(dtype(int64)).map(int)



from collections.abc import Iterable
from typing import TypeVar

from beartype import beartype
from more_itertools import one as _one

from utilities.errors import redirect_error

_T = TypeVar("_T")


@beartype
def one(iterable: Iterable[_T], /) -> _T:
    """Return the unique item from `iterable`."""
    try:
        return _one(iterable)
    except ValueError as error1:
        try:
            redirect_error(
                error1,
                "too few items in iterable",
                EmptyIterableError(error1.args[0]),
            )
        except ValueError as error2:
            redirect_error(
                error1,
                "Expected exactly one item in iterable",
                MultipleElementsError(error2.args[0]),
            )


class EmptyIterableError(Exception):
    """Raised when an iterable is empty."""


class MultipleElementsError(Exception):
    """Raised when an iterable contains multiple elements."""



from enum import auto

from hypothesis import given
from hypothesis.strategies import DataObject
from hypothesis.strategies import data
from hypothesis.strategies import sampled_from

from utilities.enum import StrEnum
from utilities.enum import parse_enum


class Truth(StrEnum):
    true = auto()
    false = auto()


class TestParseEnum:
    @given(data=data(), truth=sampled_from(Truth))
    def test_main(self, data: DataObject, truth: Truth) -> None:
        input_ = data.draw(sampled_from([truth, truth.name]))
        result = parse_enum(Truth, input_)
        assert result is truth


class TestStrEnum:
    @given(truth=sampled_from(Truth))
    def test_main(self, truth: Truth) -> None:
        assert isinstance(truth, str)
        assert truth == truth.name



from collections.abc import Iterable
from collections.abc import Iterator
from logging import getLogger
from pathlib import Path

from beartype import beartype
from click import command
from click import option

from utilities.logging import basic_config
from utilities.re import NoMatchesError
from utilities.re import extract_group
from utilities.rotate_logs.classes import Config
from utilities.rotate_logs.classes import Item

_LOGGER = getLogger(__name__)


@command()
@option("-p", "--path", type=Path, default=Config.path, show_default=True)
@option(
    "-e",
    "--extension",
    type=str,
    default=Config.extension,
    show_default=True,
)
@option("-s", "--size", type=int, default=Config.size, show_default=True)
@option("-k", "--keep", type=int, default=Config.keep, show_default=True)
@option("-dr", "--dry-run", is_flag=True, show_default=True)
@beartype
def main(
    *,
    path: Path,
    extension: str,
    size: int,
    keep: int,
    dry_run: bool,
) -> None:
    """CLI for the `rotate_logs` script."""
    basic_config()

    # log config
    _LOGGER.info("path      = %s", path)
    _LOGGER.info("extension = %s", extension)
    _LOGGER.info("size      = %s", size)
    _LOGGER.info("keep      = %s", keep)
    _LOGGER.info("dry_run   = %s", dry_run)

    # main
    item_lists = _yield_items(path=path, extension=extension, size=size)
    if dry_run:
        for items in item_lists:
            for item in items:
                _LOGGER.debug("%s", item)
    else:
        for items in item_lists:
            _rotate_items(items, keep=keep)


@beartype
def _rotate_items(items: Iterable[Item], /, *, keep: int = Config.keep) -> None:
    for item in sorted(items, key=_key, reverse=True):
        _rotate_item(item, keep=keep)


@beartype
def _key(item: Item, /) -> int:
    return item.num


@beartype
def _rotate_item(item: Item, /, *, keep: int = Config.keep) -> None:
    path = item.path
    if item.num >= keep:
        _LOGGER.info("Removing file: %s", path)
        path.unlink(missing_ok=True)
    else:
        head = item.head
        new_num = item.num + 1
        new_path = head.with_name(f"{head.name}.{new_num}")
        _LOGGER.info("Rotating file: %s -> %s", path, new_path)
        _ = path.rename(new_path)


@beartype
def _yield_items(
    *,
    path: Path = Config.path,
    extension: str = Config.extension,
    size: int = Config.size,
) -> Iterator[frozenset[Item]]:
    for p in path.rglob("*"):
        if p.suffix == f".{extension}" and p.stat().st_size >= size:
            yield frozenset(_yield_for_head(p, extension=extension))


@beartype
def _yield_for_head(
    path: Path,
    /,
    *,
    extension: str = Config.extension,
) -> Iterator[Item]:
    yield Item(path, path)
    pattern = rf"^{path.stem}\.{extension}\.(\d+)$"
    for p in path.parent.iterdir():
        try:
            num = extract_group(pattern, p.name)
        except NoMatchesError:
            pass
        else:
            yield Item(p, path, num=int(num))



from hypothesis import given
from hypothesis.strategies import DataObject
from hypothesis.strategies import data
from luigi import Task

from utilities.hypothesis.luigi import task_namespaces


class TestTaskNamespaces:
    @given(data=data())
    def test_main(self, data: DataObject) -> None:
        _ = data.draw(task_namespaces())

    @given(namespace=task_namespaces())
    def test_first(self, namespace: str) -> None:
        class Example(Task):
            task_namespace = namespace

        _ = Example()

    @given(namespace=task_namespaces())
    def test_second(self, namespace: str) -> None:
        class Example(Task):
            task_namespace = namespace
            ...

        _ = Example()



from pathlib import Path
from re import search
from typing import Optional

from hypothesis import given
from hypothesis.errors import InvalidArgument
from hypothesis.strategies import DataObject
from hypothesis.strategies import DrawFn
from hypothesis.strategies import booleans
from hypothesis.strategies import composite
from hypothesis.strategies import data
from hypothesis.strategies import integers
from hypothesis.strategies import just
from hypothesis.strategies import none
from hypothesis.strategies import sets
from pytest import mark
from pytest import param
from pytest import raises

from utilities.hypothesis import assume_does_not_raise
from utilities.hypothesis import lists_fixed_length
from utilities.hypothesis import setup_hypothesis_profiles
from utilities.hypothesis import temp_dirs
from utilities.hypothesis import temp_paths
from utilities.hypothesis import text_ascii
from utilities.hypothesis import text_clean
from utilities.hypothesis import text_printable
from utilities.tempfile import TemporaryDirectory


class TestAssumeDoesNotRaise:
    @given(x=booleans())
    def test_no_match_and_suppressed(self, x: bool) -> None:
        with assume_does_not_raise(ValueError):
            if x is True:
                msg = "x is True"
                raise ValueError(msg)
        assert x is False

    @given(x=booleans())
    def test_no_match_and_not_suppressed(self, x: bool) -> None:
        msg = "x is True"
        if x is True:
            with raises(ValueError, match=msg), assume_does_not_raise(
                RuntimeError,
            ):
                raise ValueError(msg)

    @given(x=booleans())
    def test_with_match_and_suppressed(self, x: bool) -> None:
        msg = "x is True"
        if x is True:
            with assume_does_not_raise(ValueError, match=msg):
                raise ValueError(msg)
        assert x is False

    @given(x=just(True))
    def test_with_match_and_not_suppressed(self, x: bool) -> None:
        msg = "x is True"
        if x is True:
            with raises(ValueError, match=msg), assume_does_not_raise(
                ValueError,
                match="wrong",
            ):
                raise ValueError(msg)


class TestLiftDraw:
    @given(data=data(), x=booleans())
    def test_fixed(self, data: DataObject, x: bool) -> None:
        @composite
        def func(_draw: DrawFn, /) -> bool:
            _ = _draw
            return x

        result = data.draw(func())
        assert result is x

    @given(data=data())
    def test_strategy(self, data: DataObject) -> None:
        @composite
        def func(_draw: DrawFn, /) -> bool:
            return _draw(booleans())

        result = data.draw(func())
        assert isinstance(result, bool)


class TestListsFixedLength:
    @given(data=data(), size=integers(1, 10))
    @mark.parametrize(
        "unique",
        [param(True, id="unique"), param(False, id="no unique")],
    )
    @mark.parametrize(
        "sorted_",
        [param(True, id="sorted"), param(False, id="no sorted")],
    )
    def test_main(
        self,
        data: DataObject,
        size: int,
        unique: bool,
        sorted_: bool,
    ) -> None:
        result = data.draw(
            lists_fixed_length(integers(), size, unique=unique, sorted=sorted_),
        )
        assert isinstance(result, list)
        assert len(result) == size
        if unique:
            assert len(set(result)) == len(result)
        if sorted_:
            assert sorted(result) == result


class TestSetupHypothesisProfiles:
    def test_main(self) -> None:
        setup_hypothesis_profiles()


class TestTempDirs:
    @given(temp_dir=temp_dirs())
    def test_main(self, temp_dir: TemporaryDirectory) -> None:
        _test_temp_path(temp_dir.name)

    @given(
        temp_dir=temp_dirs(),
        contents=sets(text_ascii(min_size=1), max_size=10),
    )
    def test_writing_files(
        self,
        temp_dir: TemporaryDirectory,
        contents: set[str],
    ) -> None:
        _test_writing_to_temp_path(temp_dir.name, contents)


class TestTempPaths:
    @given(temp_path=temp_paths())
    def test_main(self, temp_path: Path) -> None:
        _test_temp_path(temp_path)

    @given(
        temp_path=temp_paths(),
        contents=sets(text_ascii(min_size=1), max_size=10),
    )
    def test_writing_files(self, temp_path: Path, contents: set[str]) -> None:
        _test_writing_to_temp_path(temp_path, contents)


def _test_temp_path(path: Path, /) -> None:
    assert path.is_dir()
    assert len(set(path.iterdir())) == 0


def _test_writing_to_temp_path(path: Path, contents: set[str], /) -> None:
    assert len(set(path.iterdir())) == 0
    for content in contents:
        path.joinpath(content).touch()
    assert len(set(path.iterdir())) == len(contents)


class TestTextAscii:
    @given(
        data=data(),
        min_size=integers(0, 100),
        max_size=integers(0, 100) | none(),
    )
    def test_main(
        self,
        data: DataObject,
        min_size: int,
        max_size: Optional[int],
    ) -> None:
        with assume_does_not_raise(InvalidArgument, AssertionError):
            text = data.draw(text_ascii(min_size=min_size, max_size=max_size))
        assert search("^[A-Za-z]*$", text)
        assert len(text) >= min_size
        if max_size is not None:
            assert len(text) <= max_size


class TestTextClean:
    @given(
        data=data(),
        min_size=integers(0, 100),
        max_size=integers(0, 100) | none(),
    )
    def test_main(
        self,
        data: DataObject,
        min_size: int,
        max_size: Optional[int],
    ) -> None:
        with assume_does_not_raise(InvalidArgument, AssertionError):
            text = data.draw(text_clean(min_size=min_size, max_size=max_size))
        assert search("^\\S[^\\r\\n]*$|^$", text)
        assert len(text) >= min_size
        if max_size is not None:
            assert len(text) <= max_size


class TestTextPrintable:
    @given(
        data=data(),
        min_size=integers(0, 100),
        max_size=integers(0, 100) | none(),
    )
    def test_main(
        self,
        data: DataObject,
        min_size: int,
        max_size: Optional[int],
    ) -> None:
        with assume_does_not_raise(InvalidArgument, AssertionError):
            text = data.draw(
                text_printable(min_size=min_size, max_size=max_size),
            )
        assert search(
            r"^[0-9A-Za-z!\"#$%&'()*+,-./:;<=>?@\[\\\]^_`{|}~\s]*$",
            text,
        )
        assert len(text) >= min_size
        if max_size is not None:
            assert len(text) <= max_size



import datetime as dt
from collections.abc import Iterator
from functools import partial
from getpass import getuser
from itertools import islice
from logging import getLogger
from pathlib import Path
from shutil import rmtree
from typing import Optional

from beartype import beartype
from click import command
from click import option

from utilities.clean_dir.classes import Config
from utilities.clean_dir.classes import Item
from utilities.datetime import UTC
from utilities.logging import basic_config

_LOGGER = getLogger(__name__)


@command()
@option("-p", "--path", type=Path, default=Config.path, show_default=True)
@option("-d", "--days", type=int, default=Config.days, show_default=True)
@option(
    "-c",
    "--chunk-size",
    type=int,
    default=Config.chunk_size,
    show_default=True,
)
@option("-dr", "--dry-run", is_flag=True, show_default=True)
@beartype
def main(
    *,
    path: Path,
    days: int,
    chunk_size: Optional[int],
    dry_run: bool,
) -> None:
    """CLI for the `clean_dir` script."""
    basic_config()
    config = Config(
        path=path,
        days=days,
        chunk_size=chunk_size,
        dry_run=dry_run,
    )
    _log_config(config)
    if config.dry_run:
        for item in _yield_items(config):
            _LOGGER.debug("%s", item.path)
    else:
        _clean_dir(config)


@beartype
def _log_config(config: Config, /) -> None:
    _LOGGER.info("path        = %s", config.path)
    _LOGGER.info("days        = %s", config.days)
    _LOGGER.info("chunk_sizes = %s", config.chunk_size)
    _LOGGER.info("dry_run     = %s", config.dry_run)


@beartype
def _clean_dir(config: Config, /) -> None:
    while True:
        if len(items := list(_yield_items(config))) >= 1:
            for item in items:
                item.clean()
        else:
            return


@beartype
def _yield_items(config: Config, /) -> Iterator[Item]:
    it = _yield_inner(config)
    if (chunk_size := config.chunk_size) is not None:
        return islice(it, chunk_size)
    return it


@beartype
def _yield_inner(config: Config, /) -> Iterator[Item]:
    for p in config.path.rglob("*"):
        yield from _yield_from_path(p, config)


@beartype
def _yield_from_path(path: Path, config: Config, /) -> Iterator[Item]:
    if path.is_symlink():
        yield from _yield_from_path(path.resolve(), config)
    elif _is_owned_and_relative(path, config):  # pragma: no cover
        if (path.is_file() or path.is_socket()) and _is_old(path, config):
            yield Item(path, partial(_unlink_path, path))
        elif path.is_dir() and _is_empty(path):
            yield Item(path, partial(_unlink_dir, path))


@beartype
def _is_owned_and_relative(path: Path, config: Config, /) -> bool:
    try:
        return (path.owner() == getuser()) and path.is_relative_to(config.path)
    except FileNotFoundError:  # pragma: no cover
        return False


@beartype
def _is_empty(path: Path, /) -> bool:
    return len(list(path.iterdir())) == 0


@beartype
def _is_old(path: Path, config: Config, /) -> bool:
    age = dt.datetime.now(tz=UTC) - dt.datetime.fromtimestamp(
        path.stat().st_mtime,
        tz=UTC,
    )
    return age >= dt.timedelta(days=config.days)


@beartype
def _unlink_path(path: Path, /) -> None:
    _LOGGER.info("Removing file:      %s", path)
    path.unlink(missing_ok=True)


@beartype
def _unlink_dir(path: Path, /) -> None:
    _LOGGER.info("Removing directory: %s", path)
    rmtree(path)



from re import search
from typing import NoReturn
from typing import Union

from beartype import beartype

from utilities.text import ensure_str


@beartype
def redirect_error(
    old: Exception,
    pattern: str,
    new: Union[Exception, type[Exception]],
    /,
) -> NoReturn:
    """Redirect an error if a matching string is found."""
    args = old.args
    try:
        (msg,) = args
    except ValueError:
        raise NoUniqueArgError(args) from None
    else:
        if search(pattern, ensure_str(msg)):
            raise new from None
        raise old


class NoUniqueArgError(ValueError):
    """Raised when no unique argument can be found."""



from typing import Any

from pytest import mark
from pytest import param

from utilities.iterables import is_iterable_not_str


class TestIsIterableNotStr:
    @mark.parametrize(
        ("x", "expected"),
        [
            param(None, False),
            param([], True),
            param((), True),
            param("", False),
        ],
    )
    def test_main(self, x: Any, expected: bool) -> None:
        assert is_iterable_not_str(x) is expected



from pathlib import Path
from typing import Any
from typing import Optional
from typing import cast

from hypothesis import given
from hypothesis.strategies import DataObject
from hypothesis.strategies import data
from hypothesis.strategies import fixed_dictionaries
from hypothesis.strategies import integers
from hypothesis.strategies import lists
from hypothesis.strategies import none
from hypothesis.strategies import sampled_from
from hypothesis.strategies import sets
from hypothesis_sqlalchemy.sample import table_records_lists
from pytest import mark
from pytest import param
from pytest import raises
from sqlalchemy import Column
from sqlalchemy import Integer
from sqlalchemy import MetaData
from sqlalchemy import Table
from sqlalchemy import func
from sqlalchemy import insert
from sqlalchemy import select
from sqlalchemy import create_engine as _create_engine
from sqlalchemy.engine import Connection
from sqlalchemy.engine import Engine
from sqlalchemy.exc import OperationalError
from sqlalchemy.orm import declarative_base

from utilities.hypothesis import temp_paths
from utilities.hypothesis.sqlalchemy import sqlite_engines
from utilities.sqlalchemy import columnwise_max
from utilities.sqlalchemy import columnwise_min
from utilities.sqlalchemy import create_engine
from utilities.sqlalchemy import ensure_table_created
from utilities.sqlalchemy import ensure_table_dropped
from utilities.sqlalchemy import get_column_names
from utilities.sqlalchemy import get_columns
from utilities.sqlalchemy import get_dialect
from utilities.sqlalchemy import get_table
from utilities.sqlalchemy import get_table_name
from utilities.sqlalchemy import yield_connection
from utilities.sqlalchemy import yield_in_clause_rows


class TestColumnwiseMinMax:
    @given(
        values=lists(
            fixed_dictionaries(
                {
                    "x": integers(0, 100) | none(),
                    "y": integers(0, 100) | none(),
                },
            ),
            min_size=1,
            max_size=10,
        ),
        engine=sqlite_engines(),
    )
    def test_main(
        self,
        values: list[dict[str, Optional[int]]],
        engine: Engine,
    ) -> None:
        table = Table(
            "example",
            MetaData(),
            Column("id_", Integer, primary_key=True, autoincrement=True),
            Column("x", Integer),
            Column("y", Integer),
        )
        ensure_table_created(table, engine)
        with engine.begin() as conn:
            _ = conn.execute(table.insert(), values)
            res = conn.execute(table.select()).all()

        sel = select(
            table.c.x,
            table.c.y,
            columnwise_min(table.c.x, table.c.y).label("min_xy"),
            columnwise_max(table.c.x, table.c.y).label("max_xy"),
        )
        with engine.begin() as conn:
            res = conn.execute(sel).all()

        assert len(res) == len(values)
        for x, y, min_xy, max_xy in res:
            if (x is None) and (y is None):
                assert min_xy is None
                assert max_xy is None
            elif (x is not None) and (y is None):
                assert min_xy == x
                assert max_xy == x
            elif (x is None) and (y is not None):
                assert min_xy == y
                assert max_xy == y
            else:
                assert min_xy == min(x, y)
                assert max_xy == max(x, y)

    @given(engine=sqlite_engines())
    def test_label(self, engine: Engine) -> None:
        table = Table(
            "example",
            MetaData(),
            Column("id_", Integer, primary_key=True, autoincrement=True),
            Column("x", Integer),
        )
        ensure_table_created(table, engine)

        sel = select(columnwise_min(table.c.x, table.c.x))
        with engine.begin() as conn:
            _ = conn.execute(sel).all()


class TestCreateEngine:
    @given(temp_path=temp_paths())
    def test_main(self, temp_path: Path) -> None:
        engine = create_engine("sqlite", database=temp_path.name)
        assert isinstance(engine, Engine)


class TestEnsureTableCreated:
    @given(engine=sqlite_engines())
    @mark.parametrize("runs", [param(1), param(2)])
    def test_core(self, engine: Engine, runs: int) -> None:
        table = Table(
            "example",
            MetaData(),
            Column("id_", Integer, primary_key=True),
        )
        self._run_test(table, engine, runs)

    @given(engine=sqlite_engines())
    @mark.parametrize("runs", [param(1), param(2)])
    def test_orm(self, engine: Engine, runs: int) -> None:
        class Example(cast(Any, declarative_base())):
            __tablename__ = "example"

            id_ = Column(Integer, primary_key=True)

        self._run_test(Example, engine, runs)

    def _run_test(
        self,
        table_or_model: Any,
        engine: Engine,
        runs: int,
        /,
    ) -> None:
        sel = get_table(table_or_model).select()
        with raises(
            OperationalError,
            match="no such table",
        ), engine.begin() as conn:
            _ = conn.execute(sel).all()

        for _ in range(runs):
            ensure_table_created(table_or_model, engine)

        with engine.begin() as conn:
            _ = conn.execute(sel).all()


class TestEnsureTableDropped:
    @given(engine=sqlite_engines())
    @mark.parametrize("runs", [param(1), param(2)])
    def test_core(self, engine: Engine, runs: int) -> None:
        table = Table(
            "example",
            MetaData(),
            Column("id_", Integer, primary_key=True),
        )
        self._run_test(table, engine, runs)

    @given(engine=sqlite_engines())
    @mark.parametrize("runs", [param(1), param(2)])
    def test_orm(self, engine: Engine, runs: int) -> None:
        class Example(cast(Any, declarative_base())):
            __tablename__ = "example"

            id_ = Column(Integer, primary_key=True)

        self._run_test(Example, engine, runs)

    def _run_test(
        self,
        table_or_model: Any,
        engine: Engine,
        runs: int,
        /,
    ) -> None:
        table = get_table(table_or_model)
        sel = table.select()
        with engine.begin() as conn:
            table.create(conn)
            _ = conn.execute(sel).all()

        for _ in range(runs):
            ensure_table_dropped(table_or_model, engine)

        with raises(
            OperationalError,
            match="no such table",
        ), engine.begin() as conn:
            _ = conn.execute(sel).all()


class TestGetColumnNames:
    def test_core(self) -> None:
        table = Table(
            "example",
            MetaData(),
            Column("id_", Integer, primary_key=True),
        )
        self._run_test(table)

    def test_orm(self) -> None:
        class Example(cast(Any, declarative_base())):
            __tablename__ = "example"

            id_ = Column(Integer, primary_key=True)

        self._run_test(Example)

    def _run_test(self, table_or_model: Any, /) -> None:
        assert get_column_names(table_or_model) == ["id_"]


class TestGetColumns:
    def test_core(self) -> None:
        table = Table(
            "example",
            MetaData(),
            Column("id_", Integer, primary_key=True),
        )
        self._run_test(table)

    def test_orm(self) -> None:
        class Example(cast(Any, declarative_base())):
            __tablename__ = "example"

            id_ = Column(Integer, primary_key=True)

        self._run_test(Example)

    def _run_test(self, table_or_model: Any, /) -> None:
        columns = get_columns(table_or_model)
        assert isinstance(columns, list)
        assert len(columns) == 1
        assert isinstance(columns[0], Column)


class TestGetDialect:
    @given(engine=sqlite_engines())
    def test_sqlite(self, engine: Engine) -> None:
        assert get_dialect(engine) == "sqlite"

    @mark.parametrize(
        ("url", "expected"),
        [
            param("mssql+pyodbc://scott:tiger@mydsn", "mssql"),
            param("mysql://scott:tiger@localhost/foo", "mysql"),
            param("oracle://scott:tiger@127.0.0.1:1521/sidname", "oracle"),
            param(
                "postgresql://scott:tiger@localhost/mydatabase",
                "postgresql",
            ),
        ],
    )
    def test_non_sqlite(self, url: str, expected: str) -> None:
        assert get_dialect(_create_engine(url)) == expected


class TestGetTable:
    def test_core(self) -> None:
        table = Table(
            "example",
            MetaData(),
            Column("id_", Integer, primary_key=True),
        )
        result = get_table(table)
        assert result is table

    def test_orm(self) -> None:
        class Example(cast(Any, declarative_base())):
            __tablename__ = "example"

            id_ = Column(Integer, primary_key=True)

        table = get_table(Example)
        result = get_table(table)
        assert result is Example.__table__


class TestGetTableName:
    def test_core(self) -> None:
        table = Table(
            "example",
            MetaData(),
            Column("id_", Integer, primary_key=True),
        )
        result = get_table_name(table)
        expected = "example"
        assert result == expected

    def test_orm(self) -> None:
        class Example(cast(Any, declarative_base())):
            __tablename__ = "example"

            id_ = Column(Integer, primary_key=True)

        result = get_table_name(Example)
        expected = "example"
        assert result == expected


class TestYieldConnection:
    @given(engine=sqlite_engines())
    def test_engine(self, engine: Engine) -> None:
        with yield_connection(engine) as conn:
            assert isinstance(conn, Connection)

    @given(engine=sqlite_engines())
    def test_connection(self, engine: Engine) -> None:
        with engine.begin() as conn1, yield_connection(conn1) as conn2:
            assert isinstance(conn2, Connection)


class TestYieldInClauseRows:
    @given(
        data=data(),
        engine=sqlite_engines(),
        chunk_size=integers(1, 10) | none(),
    )
    def test_main(
        self,
        data: DataObject,
        engine: Engine,
        chunk_size: Optional[int],
    ) -> None:
        table = Table(
            "example",
            MetaData(),
            Column("id", Integer, primary_key=True),
        )
        rows = data.draw(table_records_lists(table, min_size=1))
        num_rows = len(rows)
        with engine.begin() as conn:
            table.create(conn)
            _ = conn.execute(insert(table).values(rows))
            assert (
                conn.execute(select([func.count()]).select_from(table)).scalar()
                == num_rows
            )
        row_vals = [row[0] for row in rows]
        values = data.draw(sets(sampled_from(row_vals)))
        result = list(
            yield_in_clause_rows(
                select(table.c.id),
                table.c.id,
                values,
                engine,
                chunk_size=chunk_size,
            ),
        )
        assert len(result) == len(values)



from collections.abc import Callable
from collections.abc import Iterable
from collections.abc import Mapping
from functools import partial
from io import StringIO
from io import TextIOWrapper
from multiprocessing import cpu_count
from typing import Any
from typing import Literal
from typing import Optional
from typing import TypeVar
from typing import Union
from typing import cast

from beartype import beartype
from pqdm import processes

from utilities.tqdm import _DEFAULTS
from utilities.tqdm import tqdm

_T = TypeVar("_T")


@beartype
def pmap(
    func: Callable[..., _T],
    /,
    *iterables: Iterable[Any],
    parallelism: Literal["processes", "threads"] = "processes",
    n_jobs: Optional[int] = None,
    bounded: bool = False,
    exception_behaviour: Literal["ignore", "immediate", "deferred"] = "ignore",
    desc: Optional[str] = _DEFAULTS.desc,
    total: Optional[Union[int, float]] = _DEFAULTS.total,
    leave: Optional[bool] = _DEFAULTS.leave,
    file: Optional[Union[TextIOWrapper, StringIO]] = _DEFAULTS.file,
    ncols: Optional[int] = _DEFAULTS.ncols,
    mininterval: Optional[float] = _DEFAULTS.mininterval,
    maxinterval: Optional[float] = _DEFAULTS.maxinterval,
    miniters: Optional[Union[int, float]] = _DEFAULTS.miniters,
    ascii: Union[bool, Optional[str]] = _DEFAULTS.ascii,  # noqa: A002
    unit: Optional[str] = _DEFAULTS.unit,
    unit_scale: Union[bool, int, Optional[str]] = _DEFAULTS.unit_scale,
    dynamic_ncols: Optional[bool] = _DEFAULTS.dynamic_ncols,
    smoothing: Optional[float] = _DEFAULTS.smoothing,
    bar_format: Optional[str] = _DEFAULTS.bar_format,
    initial: Optional[Union[int, float]] = 0,
    position: Optional[int] = _DEFAULTS.position,
    postfix: Optional[Mapping[str, Any]] = _DEFAULTS.postfix,
    unit_divisor: Optional[float] = _DEFAULTS.unit_divisor,
    write_bytes: Optional[bool] = _DEFAULTS.write_bytes,
    lock_args: Optional[tuple[Any, ...]] = _DEFAULTS.lock_args,
    nrows: Optional[int] = _DEFAULTS.nrows,
    colour: Optional[str] = _DEFAULTS.colour,
    delay: Optional[float] = _DEFAULTS.delay,
    gui: Optional[bool] = _DEFAULTS.gui,
    **kwargs: Any,
) -> list[_T]:
    """Parallel map, powered by `pqdm`."""
    return pstarmap(
        func,
        zip(*iterables),
        parallelism=parallelism,
        n_jobs=n_jobs,
        bounded=bounded,
        exception_behaviour=exception_behaviour,
        desc=desc,
        total=total,
        leave=leave,
        file=file,
        ncols=ncols,
        mininterval=mininterval,
        maxinterval=maxinterval,
        miniters=miniters,
        ascii=ascii,
        unit=unit,
        unit_scale=unit_scale,
        dynamic_ncols=dynamic_ncols,
        smoothing=smoothing,
        bar_format=bar_format,
        initial=initial,
        position=position,
        postfix=postfix,
        unit_divisor=unit_divisor,
        write_bytes=write_bytes,
        lock_args=lock_args,
        nrows=nrows,
        colour=colour,
        delay=delay,
        gui=gui,
        **kwargs,
    )


@beartype
def pstarmap(
    func: Callable[..., _T],
    iterable: Iterable[tuple[Any, ...]],
    /,
    *,
    parallelism: Literal["processes", "threads"] = "processes",
    n_jobs: Optional[int] = None,
    bounded: bool = False,
    exception_behaviour: Literal["ignore", "immediate", "deferred"] = "ignore",
    desc: Optional[str] = _DEFAULTS.desc,
    total: Optional[Union[int, float]] = _DEFAULTS.total,
    leave: Optional[bool] = _DEFAULTS.leave,
    file: Optional[Union[TextIOWrapper, StringIO]] = _DEFAULTS.file,
    ncols: Optional[int] = _DEFAULTS.ncols,
    mininterval: Optional[float] = _DEFAULTS.mininterval,
    maxinterval: Optional[float] = _DEFAULTS.maxinterval,
    miniters: Optional[Union[int, float]] = _DEFAULTS.miniters,
    ascii: Union[bool, Optional[str]] = _DEFAULTS.ascii,  # noqa: A002
    unit: Optional[str] = _DEFAULTS.unit,
    unit_scale: Union[bool, int, Optional[str]] = _DEFAULTS.unit_scale,
    dynamic_ncols: Optional[bool] = _DEFAULTS.dynamic_ncols,
    smoothing: Optional[float] = _DEFAULTS.smoothing,
    bar_format: Optional[str] = _DEFAULTS.bar_format,
    initial: Optional[Union[int, float]] = 0,
    position: Optional[int] = _DEFAULTS.position,
    postfix: Optional[Mapping[str, Any]] = _DEFAULTS.postfix,
    unit_divisor: Optional[float] = _DEFAULTS.unit_divisor,
    write_bytes: Optional[bool] = _DEFAULTS.write_bytes,
    lock_args: Optional[tuple[Any, ...]] = _DEFAULTS.lock_args,
    nrows: Optional[int] = _DEFAULTS.nrows,
    colour: Optional[str] = _DEFAULTS.colour,
    delay: Optional[float] = _DEFAULTS.delay,
    gui: Optional[bool] = _DEFAULTS.gui,
    **kwargs: Any,
) -> list[_T]:
    """Parallel starmap, powered by `pqdm`."""
    n_jobs = _get_n_jobs(n_jobs)
    tqdm_class = cast(Any, tqdm)
    if parallelism == "processes":
        result = processes.pqdm(
            iterable,
            partial(_starmap_helper, func),
            n_jobs=n_jobs,
            argument_type="args",
            bounded=bounded,
            exception_behaviour=exception_behaviour,
            tqdm_class=tqdm_class,
            **({} if desc is None else {"desc": desc}),
            total=total,
            leave=leave,
            file=file,
            ncols=ncols,
            mininterval=mininterval,
            maxinterval=maxinterval,
            miniters=miniters,
            ascii=ascii,
            unit=unit,
            unit_scale=unit_scale,
            dynamic_ncols=dynamic_ncols,
            smoothing=smoothing,
            bar_format=bar_format,
            initial=initial,
            position=position,
            postfix=postfix,
            unit_divisor=unit_divisor,
            write_bytes=write_bytes,
            lock_args=lock_args,
            nrows=nrows,
            colour=colour,
            delay=delay,
            gui=gui,
            **kwargs,
        )
    else:
        result = processes.pqdm(
            iterable,
            partial(_starmap_helper, func),
            n_jobs=n_jobs,
            argument_type="args",
            bounded=bounded,
            exception_behaviour=exception_behaviour,
            tqdm_class=tqdm_class,
            **({} if desc is None else {"desc": desc}),
            total=total,
            leave=leave,
            file=file,
            ncols=ncols,
            mininterval=mininterval,
            maxinterval=maxinterval,
            miniters=miniters,
            ascii=ascii,
            unit=unit,
            unit_scale=unit_scale,
            dynamic_ncols=dynamic_ncols,
            smoothing=smoothing,
            bar_format=bar_format,
            initial=initial,
            position=position,
            postfix=postfix,
            unit_divisor=unit_divisor,
            write_bytes=write_bytes,
            lock_args=lock_args,
            nrows=nrows,
            colour=colour,
            delay=delay,
            gui=gui,
            **kwargs,
        )
    return list(result)


@beartype
def _get_n_jobs(n_jobs: Optional[int], /) -> int:
    if (n_jobs is None) or (n_jobs <= 0):
        return cpu_count()  # pragma: no cover
    return n_jobs


@beartype
def _starmap_helper(func: Callable[..., _T], *args: Any) -> _T:
    return func(*args)
